from rest_framework.response import Response
from rest_framework.generics import GenericAPIView, ListAPIView

from spillway import filters, forms, renderers
from spillway.generics import BaseGeoView


class MapView(GenericAPIView):
    """View for rendering map tiles from /{z}/{x}/{y}/ tile coordinates."""
    renderer_classes = (renderers.MapnikRenderer, renderers.MapnikJPEGRenderer)

    def get_renderer_context(self):
        context = super(MapView, self).get_renderer_context()
        form = forms.MapTile(dict(self.request.query_params.dict(),
                                  **context.pop('kwargs')))
        context.update(form.cleaned_data if form.is_valid() else {})
        return context

    def get(self, request, *args, **kwargs):
        return Response(self.get_object())


class TileView(BaseGeoView, ListAPIView):
    """View for serving tiled GeoJSON from a GeoModel."""
    pagination_class = None
    filter_backends = (filters.TileFilter,)
    renderer_classes = (renderers.GeoJSONRenderer,)



"""Serializer fields"""
import os

from django.contrib.gis import forms
from rest_framework.fields import Field, FileField
from greenwich.io import MemFileIO
from greenwich.raster import Raster, driver_for_path

from spillway.compat import json


class GeometryField(Field):
    def to_internal_value(self, data):
        # forms.GeometryField cannot handle geojson dicts.
        if isinstance(data, dict):
            data = json.dumps(data)
        return forms.GeometryField().to_python(data)

    def to_representation(self, value):
        # Create a dict from the GEOSGeometry when the value is not previously
        # serialized from the spatial db.
        try:
            return {'type': value.geom_type, 'coordinates': value.coords}
        # Value is already serialized as geojson, kml, etc.
        except AttributeError:
            return value


class NDArrayField(FileField):
    def to_representation(self, value):
        geom = self.context.get('g')
        stat = self.context.get('stat')
        with Raster(getattr(value, 'path', value)) as r:
            if geom:
                with r.clip(geom) as clipped:
                    arr = clipped.masked_array()
            else:
                arr = r.masked_array()
        return arr if not stat else getattr(arr, stat)()


class GDALField(FileField):
    def to_representation(self, value):
        imgpath = value.path
        geom = self.context.get('g')
        # Handle format as .tif, tif, or tif.zip
        ext = self.context.get('format') or os.path.splitext(imgpath)[-1][1:]
        ext = os.path.splitext(ext)[0]
        # No conversion is needed if the original format without clipping
        # is requested.
        if not geom and imgpath.endswith(ext):
            return imgpath
        driver = driver_for_path('base.%s' % ext)
        memio = MemFileIO()
        if geom:
            with Raster(imgpath) as r:
                with r.clip(geom) as clipped:
                    clipped.save(memio, driver)
        else:
            driver.copy(imgpath, memio.name)
        return memio



from rest_framework.exceptions import ValidationError


class ModelSerializerMixin(object):
    """Provides generic model serializer classes to views."""
    model_serializer_class = None

    def get_serializer_class(self):
        if self.serializer_class:
            return self.serializer_class
        class DefaultSerializer(self.model_serializer_class):
            class Meta:
                model = self.queryset.model
        return DefaultSerializer


class QueryFormMixin(object):
    """Provides form based handling of GET or POST requests."""
    query_form_class = None

    def clean_params(self):
        """Returns a validated form dict from Request parameters."""
        form = self.query_form_class(
            self.request.query_params or self.request.data,
            self.request.FILES or None)
        if form.is_valid():
            return form.cleaned_data
        raise ValidationError(form.errors)



import os
import datetime

from django.contrib.gis.db import models
from django.utils.deconstruct import deconstructible
from django.utils.translation import ugettext_lazy as _
import greenwich
import numpy as np

from spillway.compat import mapnik


# Workaround for migrations and FileField upload_to, see:
# https://code.djangoproject.com/ticket/22999
@deconstructible
class UploadDir(object):
    def __init__(self, path):
        self.path = path

    def __call__(self, instance, filename):
        return os.path.join(self.path, filename)

upload_to = UploadDir('data')


class AbstractRasterStore(models.Model):
    """Abstract model for raster data storage."""
    image = models.FileField(_('raster file'), upload_to=upload_to)
    width = models.IntegerField(_('width in pixels'))
    height = models.IntegerField(_('height in pixels'))
    geom = models.PolygonField(_('raster bounding polygon'))
    event = models.DateField()
    srs = models.CharField(_('spatial reference system'), max_length=256)
    minval = models.FloatField(_('minimum value'))
    maxval = models.FloatField(_('maximum value'))
    nodata = models.FloatField(_('nodata value'), blank=True, null=True)
    # Spatial resolution
    xpixsize = models.FloatField(_('West to East pixel resolution'))
    ypixsize = models.FloatField(_('North to South pixel resolution'))

    class Meta:
        unique_together = ('image', 'event')
        ordering = ['image']
        get_latest_by = 'event'
        abstract = True

    def __unicode__(self):
        return self.image.name

    def bin(self, k=5, quantiles=False):
        if not quantiles:
            return np.linspace(self.minval, self.maxval, k)
        with greenwich.Raster(self.image.path) as rast:
            arr = rast.masked_array()
        q = list(np.linspace(0, 100, k))
        return np.percentile(arr.compressed(), q)

    def clean_fields(self, *args, **kwargs):
        # Override this instead of save() so that fields are populated on
        # save() *or* manager methods like RasterStore.objects.create().
        if not self.image.storage.exists(self.image):
            self.image.save(self.image.name, self.image, save=False)
        with greenwich.Raster(self.image.path) as r:
            band = r[-1]
            bmin, bmax = band.GetMinimum(), band.GetMaximum()
            if bmin is None or bmax is None:
                bmin, bmax = band.ComputeRasterMinMax()
            self.geom = buffer(r.envelope.polygon.ExportToWkb())
            if r.sref.srid:
                self.geom.srid = r.sref.srid
            self.xpixsize, self.ypixsize = r.affine.scale
            self.width, self.height = r.size
            self.minval = bmin
            self.maxval = bmax
            self.nodata = r.nodata
            self.srs = r.sref.wkt
        if self.event is None:
            self.event = datetime.date.today()
        super(AbstractRasterStore, self).clean_fields(*args, **kwargs)

    def save(self, *args, **kwargs):
        self.full_clean()
        super(AbstractRasterStore, self).save(*args, **kwargs)

    def layer(self, band=1):
        layer = mapnik.Layer(
            str(self), greenwich.SpatialReference(self.srs).proj4)
        layer.datasource = mapnik.Gdal(file=self.image.path, band=band)
        return layer



from rest_framework import pagination
from rest_framework.response import Response

from spillway import query
from spillway.collections import NamedCRS


class FeaturePagination(pagination.PageNumberPagination):
    """Feature pagination by page number."""

    def get_paginated_response(self, data):
        paginator = self.page.paginator
        crs = NamedCRS(query.get_srid(paginator.object_list))
        data.update({'count': paginator.count,
                     'next': self.get_next_link(),
                     'previous': self.get_previous_link(),
                     'crs': crs})
        return Response(data)



from rest_framework import viewsets, generics, mixins

from spillway.generics import BaseGeoView


class GenericGeoViewSet(BaseGeoView,
                        viewsets.ViewSetMixin,
                        generics.GenericAPIView):
    """A generic view set providing vector geometry representations."""


class ReadOnlyGeoModelViewSet(mixins.RetrieveModelMixin,
                              mixins.ListModelMixin,
                              GenericGeoViewSet):
    """A geo-enabled view set with default list and retrieve actions."""


class GeoModelViewSet(mixins.CreateModelMixin,
                      mixins.RetrieveModelMixin,
                      mixins.UpdateModelMixin,
                      mixins.DestroyModelMixin,
                      mixins.ListModelMixin,
                      GenericGeoViewSet):
    """A geo-enabled view set with default create, retrieve, update,
    partial_update, destroy, and list actions.
    """



from spillway.forms.fields import (
    BoundingBoxField, CommaSepFloatField,
    OGRGeometryField, SpatialReferenceField)
from spillway.forms.forms import (
    GeometryQueryForm, MapTile, RasterQueryForm, SpatialQueryForm)



import math

from django.contrib.gis import gdal, forms
from greenwich.srs import transform_tile

from spillway import query, styles
from spillway.compat import ALL_TERMS
from spillway.forms import fields


class GeoQuerySetForm(forms.Form):
    """Base form for applying GeoQuerySet methods and filters."""

    def __init__(self, data=None, queryset=None, *args, **kwargs):
        super(GeoQuerySetForm, self).__init__(data, *args, **kwargs)
        self.queryset = queryset
        self._is_selected = False

    def query(self):
        """Returns the filtered/selected GeoQuerySet."""
        if not self.is_valid():
            raise ValueError('Invalid field values')
        if not self._is_selected:
            if self.queryset is None:
                raise TypeError('Must be GeoQuerySet not %s' %
                                type(self.queryset))
            self.select()
            self._is_selected = True
        return self.queryset

    def select(self):
        """Set GeoQuerySet from field values and filters.

        Subclasses implement this. Not called directly, use .select().
        """
        raise NotImplementedError


class SpatialQueryForm(GeoQuerySetForm):
    """A Form for spatial lookup queries such as intersects, overlaps, etc.

    Includes 'bbox' as an alias for 'bboverlaps'.
    """
    bbox = fields.BoundingBoxField(required=False)

    def __init__(self, *args, **kwargs):
        super(SpatialQueryForm, self).__init__(*args, **kwargs)
        for lookup in self.data:
            if lookup in ALL_TERMS:
                self.fields[lookup] = forms.GeometryField(required=False)
                break

    def clean(self):
        cleaned_data = super(SpatialQueryForm, self).clean()
        spatial_lookup = set(cleaned_data.keys()) - {'bbox'}
        bbox = cleaned_data.pop('bbox', None)
        # Look for "bbox" which is just an alias to "bboverlaps".
        if bbox:
            cleaned_data.pop(spatial_lookup, None)
            cleaned_data['bboverlaps'] = bbox
        return cleaned_data

    def select(self):
        self.queryset = query.filter_geometry(
            self.queryset, **self.cleaned_data)


class GeometryQueryForm(GeoQuerySetForm):
    """A form providing GeoQuerySet method arguments."""
    format = forms.CharField(required=False)
    precision = forms.IntegerField(required=False, initial=4)
    # Tolerance value for geometry simplification
    simplify = forms.FloatField(required=False)
    srs = fields.SpatialReferenceField(required=False)

    def clean_precision(self):
        # Unfortunately initial values are not used as default values.
        return (self.cleaned_data['precision'] or
                self.fields['precision'].initial)

    def select(self):
        data = self.cleaned_data
        kwargs = {'precision': data['precision']}
        tolerance, srs, format = map(data.get, ('simplify', 'srs', 'format'))
        srid = getattr(srs, 'srid', None)
        try:
            has_format = self.queryset.has_format(format)
        except AttributeError:
            # Handle default GeoQuerySet.
            try:
                self.queryset = getattr(self.queryset, format)(**kwargs)
            except AttributeError:
                pass
        else:
            if has_format:
                kwargs.update(format=format)
            self.queryset = self.queryset.simplify(tolerance, srid, **kwargs)


class RasterQueryForm(forms.Form):
    """Validates format options for raster data."""
    bbox = fields.BoundingBoxField(required=False)
    g = fields.OGRGeometryField(required=False)
    upload = fields.GeometryFileField(required=False)
    periods = forms.IntegerField(required=False)
    stat = forms.ChoiceField(
        choices=[(choice,) * 2 for choice in 'max', 'mean', 'min', 'std'],
        required=False)

    def clean(self):
        """Return cleaned fields as a dict, determine which geom takes
        precedence.
        """
        cleaned = super(RasterQueryForm, self).clean()
        cleaned['g'] = (cleaned.pop('upload') or cleaned.pop('bbox') or
                        cleaned.get('g'))
        return cleaned


class MapTile(GeoQuerySetForm):
    """Validates requested map tiling parameters."""
    bbox = fields.OGRGeometryField(srid=4326, required=False)
    clip = forms.BooleanField(required=False, initial=False)
    x = forms.IntegerField()
    y = forms.IntegerField()
    z = forms.IntegerField()
    band = forms.IntegerField(required=False, initial=1)
    size = forms.IntegerField(required=False, initial=256)
    style = forms.ChoiceField(
        choices=[(k, k.lower()) for k in list(styles.colors)],
        required=False)
    # Tile grid uses 3857, but coordinates should be in 4326 commonly.
    tile_srid = 3857
    # Geometry simplification tolerances based on tile zlevel, see
    # http://wiki.openstreetmap.org/wiki/Slippy_map_tilenames.
    #SpatialReference(3857).GetSemiMajor() == 6378137.0
    tolerances = [6378137 * 2 * math.pi / (2 ** (zoom + 8))
                  for zoom in range(20)]

    def clean(self):
        cleaned = super(MapTile, self).clean()
        x, y, z = map(cleaned.get, ('x', 'y', 'z'))
        # Create bbox from NW and SE tile corners.
        extent = transform_tile(x, y, z) + transform_tile(x + 1, y + 1, z)
        geom = gdal.OGRGeometry.from_bbox(extent)
        geom.srid = self.fields['bbox'].srid
        cleaned['bbox'] = geom
        return cleaned

    def select(self):
        data = self.cleaned_data
        bbox = data['bbox']
        geom_wkt = bbox.ewkt
        coord_srid = bbox.srid
        original_srid = self.queryset.geo_field.srid
        try:
            tolerance = self.tolerances[data['z']]
        except IndexError:
            tolerance = self.tolerances[-1]
        attrname = query.geo_field(self.queryset).name
        self.queryset = query.filter_geometry(self.queryset,
                                              intersects=geom_wkt)
        if data['clip']:
            self.queryset = self.queryset.intersection(geom_wkt)
            attrname = 'intersection'
        for obj in self.queryset:
            geom = getattr(obj, attrname)
            # Geometry must be in Web Mercator for simplification.
            if geom.srid != self.tile_srid:
                # Result of intersection does not have SRID set properly.
                if geom.srid is None:
                    geom.srid = original_srid
                geom.transform(self.tile_srid)
            geom = geom.simplify(tolerance, preserve_topology=True)
            geom.transform(coord_srid)
            obj.geojson = geom.geojson



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(name='django-spillway',
      version='0.4.3',
      description='Geodata extensions for Django REST Framework',
      long_description=open('README.rst').read(),
      author='Brian Galey',
      author_email='bkgaley@gmail.com',
      url='https://github.com/bkg/django-spillway',
      packages=find_packages(exclude=['tests*']),
      include_package_data=True,
      install_requires=['django', 'djangorestframework>=3.1.0', 'greenwich>=0.3'],
      extras_require={'mapnik': ['Mapnik>=2.0']},
      license='BSD',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: BSD License',
          'Natural Language :: English',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
      ],
      test_suite='runtests.runtests')
